<?php


/* ****************************************************************************
	INPUT PARAMETERS
	
		dataset ID	- used for dataset path
		delete flag - whether to delet old data if exists
		
		
	OUTPUTS
		
		message to log files (error or status)
		message as HTML result


    STATE == 1
        return actual info about dataset with ID if exists

    STATE == 2
        
/* **************************************************************************** */

define( "LOG_FILENAME", "htk_process.log" );
define( "CONFIG_FILENAME", "config.xml" );
define( "HMM_CONFIG_FILENAME", "HMMConfig" );
umask(0);

        require_once('../simpleQueue/init.php');
        require_once('../simpleQueue/TaskQueue.php');
		 

/* ****************************************************************************

	STATUS and ERROR MESSAGE FORMAT
	ID; message; value
	
	Status is stored for each dataset separately.
	Error log file is common for all datasets.
	
*/

define( "STATUS_START", 1 );			// value: not used
define( "STATUS_INIT", 2 );				// value: not used
define( "STATUS_RUN", 3 );				// value: percentage of processing + config
define( "STATUS_DONE", 4 );				// value: path to the result + config

define( "ERROR_NO_CONF",         -1 );		// value: not used
define( "ERROR_NO_ID",           -2 );
define( "ERROR_NO_STATUS",       -3 );
define( "ERROR_WRONG_STATUS",    -4 );
define( "ERROR_NO_DATAFILE",     -6 );
define( "ERROR_DOWNLOAD",        -7 );
define( "ERROR_UNZIP", 		     -8 );
define( "ERROR_DATASET_CLEAN",   -9 );
define( "ERROR_DATASET_CREATE", -10 );
define( "ERROR_DATASET_NOT_VALID", -11 );
define( "ERROR_WRONG_CONFIG_PARAMS", -12 );
define( "ERROR_NOT_ZIP_FILE", -13 );

define( "ERROR_STATE",         -100 );
/* **************************************************************************** */


function msg_HTML( $no, $msg, $data = "" )
{
	global $conf;
    $out  = "<div id=\"server_timestamp\">".date("Y-m-d H:m:s")."</div>";
    $out .= "<div id=\"server_status_no\">$no</div>";
    $out .= "<div id=\"server_status_msg\">$msg</div>";
    $out .= "<div id=\"server_status_data\">$data</div><br>";
	return $out;
}

function msg_LOG( $no, $msg, $data = "" )
{
    return sprintf( "%s; %s; %s; %s;\n", date("Y-m-d H:m:s"), $no, $msg, $data );
}

function error( $no, $msg, $data = "" )
{
    if( $f = @fopen( LOG_FILENAME, "at" ) )
    {
        fwrite( $f, msg_LOG( $no, $msg, $data ) );
        fclose( $f );
    }
	echo msg_HTML( $no, $msg, $data );
    exit();
} 

function status( $no, $msg, $data = "" )
{
	global $target_path;
	if( !empty( $target_path ) ) return;

	if( $f = @fopen( $target_path.STATUS_FILENAME, "at" ) )
    {
        fwrite( $f, msg_LOG( $no, $msg, $data ) );
        fclose( $f );
    }
	echo msg_HTML( $no, $msg, $data );
} 

function string_to_filename($word) 
{
    $tmp = preg_replace('/^\W+|\W+$/', '', $word); // remove all non-alphanumeric chars at begin & end of string
    $tmp = preg_replace('/\s+/', '_', $tmp); // compress internal whitespace and replace with _
    return strtolower(preg_replace('/\W-/', '', $tmp)); // remove all non-alphanumeric chars except _ and -
}



/* **************************************************************************** */

function HMMConfig_generate( $conf, $filename )
{
	global $target_path;
	
	if( !( $f = @fopen( $filename, "w" ) ) )
		error( ERROR_HMMCONFIG_GEN, "Cannot generate HMM configuration file." );
		
	$out = "#!/bin/sh
# configuration file

# width of feature vector
featureWidth=".$conf['ftrsno']."         # for example x y dx dy
automataStatesCount=".$conf['emtsno']."  # count of emiting states 
automataComplexity=".$conf['edgsno']."   # how many edges should be binded with each automata state
classCount=".$conf['clssno']."           # 

rootDir=/var/www/HTK/$target_path
labelsTraining=\$rootDir/data/labels.txt

# nasledujici promenne neni moc potreba menit:
data=\$rootDir/data

hmm0=\$rootDir/hmm0
hmm1=\$rootDir/hmm1
hmm2=\$rootDir/hmm2
hmm3=\$rootDir/hmm3 # sem se ukladaji modely
dictionary=\$rootDir/dics/dictionary # slovnik
monophones=\$rootDir/dics/monophones # seznam fonemu
prototype=\$hmm0/proto # pocatecni model 
network=\$rootDir/net/network # gramatika
latice=\$rootDir/net/latice
tmpData=\$rootDir/tmp
trainMLF=\$rootDir/mlf/train.mlf
trainScript=\$rootDir/scripts/train_htk.scp
prototype1=\$hmm1/proto
protoMMF=\$hmm1/MMF
trainMMF=\$hmm2/MMF
txt2raw=/var/www/HTK/_htk_scripts/txt2raw # FIXME
htk2lukas=\"php /var/www/HTK/_htk_scripts/HTK2lukas.php\"
status=\$rootDir/status

";
	fwrite( $f, $out );
	fclose( $f );
} 


function report_config( $conf )
{
	global $target_path;
	global $did;
		
	$out  = "<div id=\"server_report\">Report config</div>"; 
    $out .= "<div id=\"server_task_id\">".$conf['task_id']."</div>";
	$out .= "<div id=\"server_dataset_id\">$did</div>";  	
	foreach( $conf as $k => $v )
		$out .= "<div id=\"server_conf_$k\">$v</div>";
	
	$files = scandir( $target_path );
	foreach( $files as $k => $file )
		if( substr( $file, strrpos($file, '.') + 1 ) === "zip" )
		{
			$out .= "<div id=\"server_datafile\">$file</div>";  
			break;
		}
		
	echo $out;
}

// use labels.txt to validate the data
function dataset_validate()
{
	return true;
}

function check_range( $val, $min, $max ) { return ( ( $val < $min || $val > $max ) ? false : true ); }

function check_config_params( $conf )
{
	$templ = array( 'task_id', 'ftrsno', 'emtsno', 'edgsno', 'clssno');
	$ret = array_diff( $templ, array_keys( $conf ) );
//	print_r( $templ );
//	print_r( array_keys( $conf ) );
//	print_r( $ret );
	
	if( count( array_diff( array( 'task_id', 'ftrsno', 'emtsno', 'edgsno', 'clssno'), array_keys( $conf ) ) ) != 0 ) return -10; 
	
	if( !check_range( $conf['task_id'], 0, 1000000000 ) ) return -1;
	if( !check_range( $conf['ftrsno'],  1, 1000000000 ) ) return -2;
	if( !check_range( $conf['emtsno'],  1, 10 ) ) 		  return -3;
	if( !check_range( $conf['edgsno'],  1, 10 ) ) 		  return -4;
	if( !check_range( $conf['clssno'],  1, 1000 ) ) 	  return -5;

	return 1;
}

function htk_train_structure_create()
{
	global $target_path;
	global $conf;
	
    $filename = basename( $conf['files']['uploadedfile']['name'] );
	if( substr( $filename, strrpos($filename, '.') + 1 ) !== "zip" )
		error( ERROR_NOT_ZIP_FILE, "Dataset must be compressed in zip file." );
    
   	mkdir( $target_path, 0777, true );
    $filename = "input.zip";
    // upload the dataset using zip file
    if( !move_uploaded_file( $conf['files']['uploadedfile']['tmp_name'], $target_path.$filename ) ) 
        error( ERROR_DOWNLOAD, "There was an error uploading the file, please try again!" );

	chmod ( $target_path.$filename, 0666 );

/// FIXME: rozzipovani muze trvat taky docela dlouho
    // unzip the dataset 
/*
 	$zip = new ZipArchive;
	if( $zip->open( $target_path.$filename ) === TRUE) 
	{
	    $zip->extractTo( $target_path );
	    $zip->close();
	} else {
	    error( ERROR_UNZIP, "Failed to unzip ".$target_path.$filename."." );
	}


	if( !dataset_validate() )
	    error( ERROR_DATASET_NOT_VALID, "Dataset is not valid." );
*/
	mkdir( $target_path."/hmm0", 0777, true);
	mkdir( $target_path."/hmm1", 0777, true );
	mkdir( $target_path."/hmm2", 0777, true );
	mkdir( $target_path."/hmm3", 0777, true );
	mkdir( $target_path."/dics", 0777, true );
	mkdir( $target_path."/net", 0777, true );
	mkdir( $target_path."/mlf", 0777, true );
	mkdir( $target_path."/scripts", 0777, true );

    // start training process
    HMMConfig_generate( @$conf, $target_path.HMM_CONFIG_FILENAME );
    

        $queue = new TaskQueue();
 
	// add task to taskQueue (DB)
	// $task->add(...)
//	$queue->add("/var/www/HTK/_htk_scripts/HMMGen.sh /var/www/HTK/$target_path".HMM_CONFIG_FILENAME);
//	$queue->add("/var/www/HTK/_htk_scripts/HMMTrain.sh /var/www/HTK/$target_path".HMM_CONFIG_FILENAME);
	$conf['task_id'] = $queue->add("/var/www/HTK/_htk_scripts/HMMAll.sh /var/www/HTK/$target_path".HMM_CONFIG_FILENAME);

//	rand( 1, 10000 );

	// simulate progress
	$conf['progress'] = 0;

	// save_config
	file_put_contents( $target_path.CONFIG_FILENAME, xmlrpc_encode( $conf ) );
	
}

function get_process_info( $id, $data )
{
//  msg_LOG(2222, "test123: $id", "$data");
}

/* **************************************************************************** */
/* **************************************************************************** */

//print_r( $_REQUEST );
//print_r( $_FILES );


$did = @$_REQUEST['dataset_id'];
if( empty( $did ) ) error( ERROR_NO_ID, "The dataset ID must be entered." );
$did = string_to_filename( $did );

$target_path = "sessions/$did/";

$conf = @$_REQUEST['conf'];
$conf['files'] = $_FILES;
$conf['task_id'] = @$_REQUEST['task_id'];
if( empty( $conf['task_id'] ) ) $conf['task_id'] = 0;

//if( empty( $conf['task_id'] ) )			// at least task_id (def. 0) 
//	error( ERROR_NO_CONF, "Wrong input protocol." );

//print_r( $conf );

// dataset does not exists or was not completely created
if( !file_exists( $target_path ) || !file_exists( $target_path.CONFIG_FILENAME ) )
{
	if( count( $conf ) < 3 )			 
		error( STATUS_INIT, "Setup configuration parameters." );
		
    if( ( $ret  = check_config_params( $conf ) ) < 0 )
    	error( $ret, "Configuration parameters are wrong." );

     // get configuration
    $conf['datetime'] = date("Y-m-d H:m:s");
	    
   // get datafile
    if( empty($conf['files']) ) error( ERROR_NO_DATAFILE, "Select and send file with training data." );

	htk_train_structure_create();
}
else
	$conf = xmlrpc_decode( file_get_contents( $target_path.CONFIG_FILENAME ) );


// dataset exists
// get status from DB ? 
// get_info_from_db( $id, @$data );

$msg = ""; $data = "";


$queue = new TaskQueue();
$task = $queue->get($conf['task_id']);

$process_status = STATUS_RUN;

switch ($task->status) {
  case 'queued':
    $active = $queue->active_task();
    $before = (isset($active)) ? ($task->id - $active->id) : -1;
    $estimated = $queue->estimate($task->id);
    $msg = "Process is in queue";
    $data = "($before / $estimated s)";
  break;
  case 'running':
	if (file_exists($target_path."status")) {
	    $conf['progress'] = file_get_contents($target_path."status");
	}
//	if( $conf['progress'] > 100 ) $conf['progress'] = 100;
	file_put_contents( $target_path.CONFIG_FILENAME, xmlrpc_encode( $conf ) );
	$process_status = STATUS_RUN;
	$msg = "Process run info ... "; $data = $conf['progress']."% (".(time()-$task->start)."s)";
  break;
  case 'done':
	$process_status = STATUS_DONE;
	$msg = "Process results: "; 
	$data = 
	"<a href=\"".$target_path."output.zip\">".$target_path."output.zip</a> <br/>"
//	"<a href=\"".$target_path."hmm3\">".$target_path."results.zip</a> <br/>"
	  . "started: ". date("j.n.Y H:i:s", $task->start). "<br/>"
	  . "stopped: ". date("j.n.Y H:i:s", $task->stop). "<br/>"
	  . "time: ".($task->stop - $task->start). "s <br/>"
	  ." <hr/>log:<br/>".nl2br($task->stdout);
  break;
  default:
    $msg = "\$task->status ".$conf['task_id']." error: $task->status ";
  break;
}

report_config( $conf );

error( $process_status, $msg, $data );

?>  
